package com.sun.org.apache.bcel.internal.classfile;

import java.util.zip.ZipEntry;
import java.io.IOException;
import java.util.zip.ZipFile;
import java.io.FileInputStream;
import java.io.BufferedInputStream;
import java.io.InputStream;
import java.io.DataInputStream;

public final class ClassParser
{
    private DataInputStream dataInputStream;
    private final boolean fileOwned;
    private final String file_name;
    private String zip_file;
    private int class_name_index;
    private int superclass_name_index;
    private int major;
    private int minor;
    private int access_flags;
    private int[] interfaces;
    private ConstantPool constant_pool;
    private Field[] fields;
    private Method[] methods;
    private Attribute[] attributes;
    private final boolean is_zip;
    private static final int BUFSIZE = 8192;
    
    public ClassParser(InputStream inputStream, String file_name) {
        this.file_name = file_name;
        this.fileOwned = false;
        String clazz = inputStream.getClass().getName();
        this.is_zip = (clazz.startsWith("java.util.zip.") || clazz.startsWith("java.util.jar."));
        if ((inputStream instanceof DataInputStream)) {
            this.dataInputStream = (DataInputStream)inputStream;
        }
        else {
            this.dataInputStream = new DataInputStream(new BufferedInputStream(inputStream, 8192));
        }
    }
    
    public ClassParser(String file_name) {
        this.is_zip = false;
        this.file_name = file_name;
        this.fileOwned = true;
    }
    
    public ClassParser(String zip_file, String file_name) {
        this.is_zip = true;
        this.fileOwned = true;
        this.zip_file = zip_file;
        this.file_name = file_name;
    }
    
    public JavaClass parse() throws IOException, ClassFormatException {
        ZipFile zip = null;
        try {
            if (this.fileOwned) {
                if (this.is_zip) {
                    zip = new ZipFile(this.zip_file);
                    ZipEntry entry = zip.getEntry(this.file_name);
                    if (entry == null) {
                        throw new IOException("File " + this.file_name + " not found");
                    }
                    this.dataInputStream = new DataInputStream(new BufferedInputStream(zip.getInputStream(entry), 8192));
                }
                else {
                    this.dataInputStream = new DataInputStream(new BufferedInputStream(new FileInputStream(this.file_name), 8192));
                }
            }
            this.readID();
            this.readVersion();
            this.readConstantPool();
            this.readClassInfo();
            this.readInterfaces();
            this.readFields();
            this.readMethods();
            this.readAttributes();
        }
        finally {
            if (this.fileOwned) {
                try {
                    if (this.dataInputStream != null) {
                        this.dataInputStream.close();
                    }
                }
                catch (IOException ex) {}
            }
            try {
                if (zip != null) {
                    zip.close();
                }
            }
            catch (IOException ex2) {}
        }
        return new JavaClass(this.class_name_index, this.superclass_name_index, this.file_name, this.major, this.minor, this.access_flags, this.constant_pool, this.interfaces, this.fields, this.methods, this.attributes, (byte)(this.is_zip ? 3 : 2));
    }
    
    private void readAttributes() throws IOException, ClassFormatException {
        int attributes_count = this.dataInputStream.readUnsignedShort();
        this.attributes = new Attribute[attributes_count];
        for (int i = 0; i < attributes_count; ++i) {
            this.attributes[i] = Attribute.readAttribute(this.dataInputStream, this.constant_pool);
        }
    }
    
    private void readClassInfo() throws IOException, ClassFormatException {
        this.access_flags = this.dataInputStream.readUnsignedShort();
        if ((this.access_flags & 0x200) != 0x0) {
            this.access_flags |= 0x400;
        }
        if ((this.access_flags & 0x400) != 0x0 && (this.access_flags & 0x10) != 0x0) {
            throw new ClassFormatException("Class " + this.file_name + " can't be both final and abstract");
        }
        this.class_name_index = this.dataInputStream.readUnsignedShort();
        this.superclass_name_index = this.dataInputStream.readUnsignedShort();
        return;
    }
    
    private void readConstantPool() throws IOException, ClassFormatException {
        this.constant_pool = new ConstantPool(this.dataInputStream);
    }
    
    private void readFields() throws IOException, ClassFormatException {
        int fields_count = this.dataInputStream.readUnsignedShort();
        this.fields = new Field[fields_count];
        for (int i = 0; i < fields_count; ++i) {
            this.fields[i] = new Field(this.dataInputStream, this.constant_pool);
        }
    }
    
    private void readID() throws IOException, ClassFormatException {
        if (this.dataInputStream.readInt() != -889275714) {
            throw new ClassFormatException(this.file_name + " is not a Java .class file");
        }
        return;
    }
    
    private void readInterfaces() throws IOException, ClassFormatException {
        int interfaces_count = this.dataInputStream.readUnsignedShort();
        this.interfaces = new int[interfaces_count];
        for (int i = 0; i < interfaces_count; ++i) {
            this.interfaces[i] = this.dataInputStream.readUnsignedShort();
        }
    }
    
    private void readMethods() throws IOException, ClassFormatException {
        int methods_count = this.dataInputStream.readUnsignedShort();
        this.methods = new Method[methods_count];
        for (int i = 0; i < methods_count; ++i) {
            this.methods[i] = new Method(this.dataInputStream, this.constant_pool);
        }
    }
    
    private void readVersion() throws IOException, ClassFormatException {
        this.minor = this.dataInputStream.readUnsignedShort();
        this.major = this.dataInputStream.readUnsignedShort();
    }
}
